from DateTime import DateTime
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping

import datetime
import Missing

# use simplejson because it's ahead of stdlib and supports more types
import simplejson


try:
    from z3c.relationfield.relation import RelationValue
except ImportError:
    RelationValue = None


def custom_json_handler(obj):
    if obj == Missing.Value:
        return None
    obj_type = type(obj)
    if obj_type in (datetime.datetime, datetime.date):
        return obj.isoformat()
    if obj_type == DateTime:
        return obj.ISO()
    if obj_type == set:
        return list(obj)
    if obj_type == PersistentMapping:
        return dict(obj)
    if obj_type == PersistentList:
        return list(obj)
    if RelationValue is not None and obj_type == RelationValue:
        return obj.to_id
    return obj


def json_dumps(data):
    return simplejson.dumps(data, default=custom_json_handler)


# can eventually provide custom handling here if we want
json_loads = simplejson.loads



from OFS.CopySupport import _cb_encode
from OFS.CopySupport import cookie_path
from OFS.Moniker import Moniker
from plone.app.content.browser.contents import ContentsBaseAction
from plone.app.content.interfaces import IStructureAction
from plone.base import PloneMessageFactory as _
from zope.i18n import translate
from zope.interface import implementer


@implementer(IStructureAction)
class CopyAction:
    order = 2

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_options(self):
        return {
            "tooltip": translate(_("Copy"), context=self.request),
            "id": "copy",
            "icon": "plone-copy",
            "url": self.context.absolute_url() + "/@@fc-copy",
        }


class CopyActionView(ContentsBaseAction):
    success_msg = _("Successfully copied items")
    failure_msg = _("Failed to copy items")

    def action(self, obj):
        self.oblist.append(obj)

    def finish(self):
        oblist = []
        for ob in self.oblist:
            if not ob.cb_isCopyable():
                self.errors.append(
                    _(
                        "${title} cannot be copied.",
                        mapping={"title": self.objectTitle(ob)},
                    )
                )
                continue
            m = Moniker(ob)
            oblist.append(m.dump())
        cp = (0, oblist)
        cp = _cb_encode(cp)
        resp = self.request.response
        resp.setCookie("__cp", cp, path="%s" % cookie_path(self.request))
        self.request["__cp"] = cp

    def __call__(self):
        self.oblist = []
        return super().__call__(keep_selection_order=True)



from plone.app.content.testing import PLONE_APP_CONTENT_NON_ASCII_INTEGRATION_TESTING
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName

import json
import unittest


class TestNonAsciiCharactersWorkflow(unittest.TestCase):
    layer = PLONE_APP_CONTENT_NON_ASCII_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        # set non-ascii-workflow for Documents
        wf_tool = getToolByName(self, "portal_workflow")
        wf_tool.setChainForPortalTypes(["Document"], "non-ascii-workflow")

        # create an object having the non-ascii-workflow assigned
        self.portal.invokeFactory("Document", "doc")

    def test_non_ascii_characters_in_workflow_title(self):
        wf_tool = getToolByName(self, "portal_workflow")
        workflow_matching_id = list(
            filter(
                lambda workflow: ("non-ascii-workflow" in workflow.id),
                wf_tool.getWorkflowsFor(self.portal.doc),
            )
        )

        # Make sure that the non-ascii-workflow was assigned to the Document
        self.assertTrue(workflow_matching_id)

        # Build POST request to get state title for the Document
        documents_uid = IUUID(self.portal.doc)
        self.request.form["selection"] = json.dumps(documents_uid)
        self.request.form["transitions"] = True
        self.request.form["render"] = "yes"

        try:
            # try to do the json request which among other things returns
            # the workflow state title containing non-ascii characters.
            self.portal.unrestrictedTraverse("@@fc-workflow")()
        except UnicodeDecodeError:
            self.fail(
                "Calling @@fc-workflow raised UnicodeDecodeError \
                       unexpectedly."
            )



from plone.app.content.browser import vocabulary
from plone.app.content.browser.file import FileUploadView
from plone.app.content.browser.query import QueryStringIndexOptions
from plone.app.content.browser.vocabulary import VocabularyView
from plone.app.content.testing import ExampleFunctionVocabulary
from plone.app.content.testing import ExampleVocabulary
from plone.app.content.testing import PLONE_APP_CONTENT_DX_FUNCTIONAL_TESTING
from plone.app.content.testing import PLONE_APP_CONTENT_DX_INTEGRATION_TESTING
from plone.app.testing import login
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.app.z3cform.interfaces import IFieldPermissionChecker
from Products.CMFCore.indexing import processQueue
from unittest import mock
from zope.component import getMultiAdapter
from zope.component import provideAdapter
from zope.component import provideUtility
from zope.component.globalregistry import base
from zope.globalrequest import setRequest
from zope.interface import alsoProvides
from zope.interface import Interface
from zope.interface import noLongerProvides
from zope.publisher.browser import TestRequest

import json
import operator
import os
import transaction
import unittest


_dir = os.path.dirname(__file__)


class PermissionChecker:
    def __init__(self, context):
        pass

    def validate(self, field_name, vocabulary_name=None):
        if field_name == "allowed_field":
            return True
        elif field_name == "disallowed_field":
            return False
        else:
            raise AttributeError("Missing Field")


class ICustomPermissionProvider(Interface):
    pass


def _enable_permission_checker(context):
    provideAdapter(
        PermissionChecker,
        adapts=(ICustomPermissionProvider,),
        provides=IFieldPermissionChecker,
    )
    alsoProvides(context, ICustomPermissionProvider)


def _disable_permission_checker(context):
    noLongerProvides(context, ICustomPermissionProvider)
    base.unregisterAdapter(
        required=(ICustomPermissionProvider,), provided=IFieldPermissionChecker
    )


class BrowserTest(unittest.TestCase):
    layer = PLONE_APP_CONTENT_DX_INTEGRATION_TESTING

    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        setRequest(self.request)
        self.portal = self.layer["portal"]
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        provideUtility(ExampleVocabulary(), name="vocab_class")
        provideUtility(ExampleFunctionVocabulary, name="vocab_function")
        vocabulary.PERMISSIONS.update(
            {
                "vocab_class": "Modify portal content",
                "vocab_function": "Modify portal content",
            }
        )

    def testVocabularyQueryString(self):
        """Test querying a class based vocabulary with a search string."""
        view = VocabularyView(self.portal, self.request)
        self.request.form.update({"name": "vocab_class", "query": "three"})
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 1)

    def testVocabularyFunctionQueryString(self):
        """Test querying a function based vocabulary with a search string."""
        view = VocabularyView(self.portal, self.request)
        self.request.form.update({"name": "vocab_function", "query": "third"})
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 1)

    def testVocabularyNoResults(self):
        """Tests that the widgets displays correctly"""
        view = VocabularyView(self.portal, self.request)
        query = {
            "criteria": [
                {
                    "i": "path",
                    "o": "plone.app.querystring.operation.string.path",
                    "v": "/foo",
                }
            ]
        }
        self.request.form.update(
            {"name": "plone.app.vocabularies.Catalog", "query": json.dumps(query)}
        )
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 0)

    def testVocabularyCatalogResults(self):
        self.portal.invokeFactory("Document", id="page", title="page")
        self.portal.page.reindexObject()
        view = VocabularyView(self.portal, self.request)
        query = {
            "criteria": [
                {
                    "i": "path",
                    "o": "plone.app.querystring.operation.string.path",
                    "v": "/plone",
                },
                {
                    "i": "portal_type",
                    "o": "plone.app.querystring.operation.selection.any",
                    "v": "Document",
                },
            ]
        }
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Catalog",
                "query": json.dumps(query),
                "attributes": ["UID", "id", "title", "path"],
            }
        )
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 1)

    def testVocabularyCatalogUnsafeMetadataAllowed(self):
        """Users with permission "Modify portal content" are allowed to see
        ``_unsafe_metadata``.
        """
        self.portal.invokeFactory("Document", id="page", title="page")
        self.portal.page.reindexObject()
        view = VocabularyView(self.portal, self.request)
        query = {
            "criteria": [
                {
                    "i": "path",
                    "o": "plone.app.querystring.operation.string.path",
                    "v": "/plone/page",
                }
            ]
        }
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Catalog",
                "query": json.dumps(query),
                "attributes": [
                    "id",
                    "commentors",
                    "Creator",
                    "listCreators",
                ],
            }
        )
        data = json.loads(view())
        self.assertEqual(len(list(data["results"][0].keys())), 4)

    def testVocabularyCatalogUnsafeMetadataDisallowed(self):
        """Users without permission "Modify portal content" are not allowed to
        see ``_unsafe_metadata``.
        """
        self.portal.invokeFactory("Document", id="page", title="page")
        self.portal.page.reindexObject()
        # Downgrade permissions
        setRoles(self.portal, TEST_USER_ID, [])
        view = VocabularyView(self.portal, self.request)
        query = {
            "criteria": [
                {
                    "i": "path",
                    "o": "plone.app.querystring.operation.string.path",
                    "v": "/plone/page",
                }
            ]
        }
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Catalog",
                "query": json.dumps(query),
                "attributes": [
                    "id",
                    "commentors",
                    "Creator",
                    "listCreators",
                ],
            }
        )
        data = json.loads(view())
        # Only one result key should be returned, as ``commentors``,
        # ``Creator`` and ``listCreators`` is considered unsafe and thus
        # skipped.
        self.assertEqual(len(list(data["results"][0].keys())), 1)

    def testVocabularyBatching(self):
        amount = 30
        for i in range(amount):
            self.portal.invokeFactory(
                "Document", id="page" + str(i), title="Page" + str(i)
            )
            self.portal["page" + str(i)].reindexObject()
        view = VocabularyView(self.portal, self.request)
        query = {
            "criteria": [
                {
                    "i": "path",
                    "o": "plone.app.querystring.operation.string.path",
                    "v": "/plone",
                },
                {
                    "i": "portal_type",
                    "o": "plone.app.querystring.operation.selection.any",
                    "v": "Document",
                },
            ]
        }
        # batch pages are 1-based
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Catalog",
                "query": json.dumps(query),
                "attributes": ["UID", "id", "title", "path"],
                "batch": {"page": "1", "size": "10"},
            }
        )
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 10)
        self.assertEqual(data["total"], amount)

    def testVocabularyEncoding(self):
        """The vocabulary should not return the binary encoded token
        ("N=C3=A5=C3=B8=C3=AF"), but instead the value as the id in the result
        set. Fixes an encoding problem. See:
        https://github.com/plone/Products.CMFPlone/issues/650
        """
        test_val = "Nåøï"

        self.portal.invokeFactory("Document", id="page", title="page")
        self.portal.page.subject = (test_val,)
        self.portal.page.reindexObject(idxs=["Subject"])
        processQueue()

        self.request.form["name"] = "plone.app.vocabularies.Keywords"
        results = getMultiAdapter((self.portal, self.request), name="getVocabulary")()
        results = json.loads(results)
        result = results["results"][0]

        self.assertEqual(result["text"], test_val)
        self.assertEqual(result["id"], test_val)

    def testVocabularyHtmlEntity(self):
        """The vocabulary token should not convert to htmlentities.
        See https://github.com/plone/Products.CMFPlone/issues/3874
        """
        test_val = "Question & Answer"

        self.portal.invokeFactory("Document", id="page", title="page")
        self.portal.page.subject = (test_val,)
        self.portal.page.reindexObject(idxs=["Subject"])
        processQueue()

        self.request.form["name"] = "plone.app.vocabularies.Keywords"
        results = getMultiAdapter((self.portal, self.request), name="getVocabulary")()
        results = json.loads(results)
        result = results["results"][0]

        self.assertEqual(result["text"], test_val)
        self.assertEqual(result["id"], test_val)

    def testVocabularyUnauthorized(self):
        setRoles(self.portal, TEST_USER_ID, [])
        view = VocabularyView(self.portal, self.request)
        self.request.form.update(
            {"name": "plone.app.vocabularies.Users", "query": TEST_USER_NAME}
        )
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")

    def testVocabularyMissing(self):
        view = VocabularyView(self.portal, self.request)
        self.request.form.update(
            {
                "name": "vocabulary.that.does.not.exist",
            }
        )
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")

    def testPermissionCheckerAllowed(self):
        # Setup a custom permission checker on the portal
        _enable_permission_checker(self.portal)
        view = VocabularyView(self.portal, self.request)

        # Allowed field is allowed
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "allowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(
            len(data["results"]), len(self.portal.portal_types.objectIds())
        )
        _disable_permission_checker(self.portal)

    def testPermissionCheckerUnknownVocab(self):
        _enable_permission_checker(self.portal)
        view = VocabularyView(self.portal, self.request)
        # Unknown vocabulary gives error
        self.request.form.update(
            {
                "name": "vocab.does.not.exist",
                "field": "allowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(
            data["error"],
            'No factory with name "{}" exists.'.format("vocab.does.not.exist"),
        )
        _disable_permission_checker(self.portal)

    def testPermissionCheckerDisallowed(self):
        _enable_permission_checker(self.portal)
        view = VocabularyView(self.portal, self.request)
        # Disallowed field is not allowed
        # Allowed field is allowed
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "disallowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")
        _disable_permission_checker(self.portal)

    def testPermissionCheckerShortCircuit(self):
        _enable_permission_checker(self.portal)
        view = VocabularyView(self.portal, self.request)
        # Known vocabulary name short-circuits field permission check
        # global permission
        self.request.form["name"] = "plone.app.vocabularies.Users"
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Users",
                "field": "disallowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(data["results"], [])
        _disable_permission_checker(self.portal)

    def testPermissionCheckerUnknownField(self):
        _enable_permission_checker(self.portal)
        view = VocabularyView(self.portal, self.request)
        # Unknown field is raises error
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "missing_field",
            }
        )
        with self.assertRaises(AttributeError):
            view()
        _disable_permission_checker(self.portal)

    def testVocabularyUsers(self):
        acl_users = self.portal.acl_users
        membership = self.portal.portal_membership
        amount = 10
        # Let's test that safe html is used on the fullname,
        # as alternative to the workaround in PloneHotfix20210518.
        for i in range(amount):
            id = "user" + str(i)
            acl_users.userFolderAddUser(id, TEST_USER_PASSWORD, ["Member"], [])
            member = membership.getMemberById(id)
            # Make user0 the hacker.
            if i == 0:
                fullname = "user <script>alert('tag')</script> hacker"
            else:
                fullname = id
            member.setMemberProperties(mapping={"fullname": fullname})
        view = VocabularyView(self.portal, self.request)
        self.request.form.update(
            {"name": "plone.app.vocabularies.Users", "query": "user"}
        )
        data = json.loads(view())

        self.assertEqual(len(data["results"]), amount)
        # Let's sort, just to be sure.
        results = sorted(data["results"], key=operator.itemgetter("id"))
        # The first one is the hacker.  The hack should have failed.
        self.assertDictEqual(results[0], {"id": "user0", "text": "user  hacker"})

    def testSource(self):
        from z3c.form.browser.text import TextWidget
        from zope.interface import implementer
        from zope.interface import Interface
        from zope.schema import Choice
        from zope.schema.interfaces import ISource

        @implementer(ISource)
        class DummyCatalogSource:
            def search_catalog(self, query):
                querytext = query["SearchableText"]["query"]
                return [mock.Mock(id=querytext)]

        widget = TextWidget(self.request)
        widget.context = self.portal
        widget.field = Choice(source=DummyCatalogSource())
        widget.field.interface = Interface

        from plone.app.content.browser.vocabulary import SourceView

        view = SourceView(widget, self.request)
        query = {
            "criteria": [
                {
                    "i": "SearchableText",
                    "o": "plone.app.querystring.operation.string.is",
                    "v": "foo",
                }
            ]
        }
        self.request.form.update(
            {
                "query": json.dumps(query),
                "attributes": "id",
            }
        )
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["id"], "foo")

    def testSourceCollectionField(self):
        # This test uses a collection field
        # and a source providing the 'search' method
        # to help achieve coverage.
        from z3c.form.browser.text import TextWidget
        from zope.interface import implementer
        from zope.interface import Interface
        from zope.schema import Choice
        from zope.schema import List
        from zope.schema.interfaces import ISource
        from zope.schema.vocabulary import SimpleTerm

        @implementer(ISource)
        class DummySource:
            def search(self, query):
                terms = [SimpleTerm(query, query)]
                return iter(terms)

        widget = TextWidget(self.request)
        widget.context = self.portal
        widget.field = List(value_type=Choice(source=DummySource()))
        widget.field.interface = Interface

        from plone.app.content.browser.vocabulary import SourceView

        view = SourceView(widget, self.request)
        query = {
            "criteria": [
                {
                    "i": "SearchableText",
                    "o": "plone.app.querystring.operation.string.is",
                    "v": "foo",
                }
            ],
            "sort_on": "id",
            "sort_order": "ascending",
        }
        self.request.form.update(
            {
                "query": json.dumps(query),
                "batch": json.dumps({"size": 10, "page": 1}),
            }
        )
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["id"], "foo")

    def testSourcePermissionDenied(self):
        from z3c.form.browser.text import TextWidget
        from zope.interface import implementer
        from zope.interface import Interface
        from zope.schema import Choice
        from zope.schema.interfaces import ISource

        @implementer(ISource)
        class DummyCatalogSource:
            def search_catalog(self, query):
                querytext = query["SearchableText"]["query"]
                return [mock.Mock(id=querytext)]

        widget = TextWidget(self.request)
        widget.context = self.portal
        widget.field = Choice(source=DummyCatalogSource())
        widget.field.interface = Interface

        from plone.app.content.browser.vocabulary import SourceView

        view = SourceView(widget, self.request)
        query = {
            "criteria": [
                {
                    "i": "SearchableText",
                    "o": "plone.app.querystring.operation.string.is",
                    "v": "foo",
                }
            ]
        }
        self.request.form.update(
            {
                "query": json.dumps(query),
            }
        )
        logout()
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed.")

    def testSourceDefaultPermission(self):
        from plone.app.content.browser.vocabulary import SourceView
        from z3c.form.browser.text import TextWidget

        widget = TextWidget(self.request)
        view = SourceView(widget, self.request)
        self.assertEqual(view.default_permission, "cmf.ModifyPortalContent")

    def testSourceDefaultPermissionOnAddForm(self):
        from plone.app.content.browser.vocabulary import SourceView
        from z3c.form import form
        from z3c.form.browser.text import TextWidget

        widget = TextWidget(self.request)
        widget.form = form.AddForm(self.portal, self.request)

        view = SourceView(widget, self.request)
        self.assertEqual(view.default_permission, "cmf.AddPortalContent")

    def testSourceTextQuery(self):
        from z3c.form.browser.text import TextWidget
        from zope.interface import implementer
        from zope.interface import Interface
        from zope.schema import Choice
        from zope.schema.interfaces import ISource

        @implementer(ISource)
        class DummyCatalogSource:
            def search(self, query):
                return [mock.Mock(value=mock.Mock(id=query))]

        widget = TextWidget(self.request)
        widget.context = self.portal
        widget.field = Choice(source=DummyCatalogSource())
        widget.field.interface = Interface

        from plone.app.content.browser.vocabulary import SourceView

        view = SourceView(widget, self.request)
        self.request.form.update(
            {
                "query": "foo",
                "attributes": "id",
            }
        )
        data = json.loads(view())
        self.assertEqual(len(data["results"]), 1)
        self.assertEqual(data["results"][0]["id"], "foo")

    def testQueryStringConfiguration(self):
        view = QueryStringIndexOptions(self.portal, self.request)
        data = json.loads(view())
        # just test one so we know it's working...
        self.assertEqual(data["indexes"]["sortable_title"]["sortable"], True)

    @mock.patch("zope.i18n.negotiate", new=lambda ctx: "de")
    def testUntranslatableMetadata(self):
        """Test translation of ``@@getVocabulary`` view results.
        From the standard metadata columns, only ``Type`` is translated.
        """
        # Language is set via language negotiaton patch.

        self.portal.invokeFactory("Document", id="page", title="page")
        self.portal.page.reindexObject()
        view = VocabularyView(self.portal, self.request)
        query = {
            "criteria": [
                {
                    "i": "path",
                    "o": "plone.app.querystring.operation.string.path",
                    "v": "/plone/page",
                }
            ]
        }
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Catalog",
                "query": json.dumps(query),
                "attributes": [
                    "id",
                    "portal_type",
                    "Type",
                ],
            }
        )

        # data['results'] should return one item, which represents the document
        # created before.
        data = json.loads(view())

        # Type is translated
        self.assertEqual(data["results"][0]["Type"], "Seite")

        # portal_type is never translated
        self.assertEqual(data["results"][0]["portal_type"], "Document")

    def testGetMimeIcon(self):
        """Check if the returned icon is correct"""
        query = {
            "criteria": [
                {
                    "i": "portal_type",
                    "o": "plone.app.querystring.operation.selection.any",
                    "v": "File",
                },
            ]
        }
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Catalog",
                "attributes": ["getMimeIcon"],
                "query": json.dumps(query),
            }
        )
        view = VocabularyView(self.portal, self.request)

        # Check an empty file
        self.portal.invokeFactory("File", id="my-file", title="My file")
        obj = self.portal["my-file"]
        obj.reindexObject()

        self.assertListEqual(json.loads(view())["results"], [{"getMimeIcon": None}])

        # mock a pdf
        obj.file = mock.Mock(contentType="application/pdf")
        obj.reindexObject()
        self.assertListEqual(
            json.loads(view())["results"],
            [{"getMimeIcon": "/plone/++resource++mimetype.icons/pdf.png"}],
        )

        # mock something unknown
        obj.file = mock.Mock(contentType="x-foo/x-bar")
        obj.reindexObject()
        self.assertListEqual(
            json.loads(view())["results"],
            [{"getMimeIcon": "/plone/++resource++mimetype.icons/unknown.png"}],
        )


class FunctionalBrowserTest(unittest.TestCase):
    layer = PLONE_APP_CONTENT_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.request = TestRequest()
        setRequest(self.request)
        self.portal = self.layer["portal"]
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def testFileUpload(self):
        view = FileUploadView(self.portal, self.request)
        from plone.namedfile.file import FileChunk

        chunk = FileChunk(b"foobar")
        chunk.filename = "test.xml"
        self.request.form["file"] = chunk
        self.request.REQUEST_METHOD = "POST"
        # the next calls plone.app.dexterity.factories and does a
        # transaction.commit. Needs cleanup and FunctionalTesting layer.
        data = json.loads(view())
        self.assertEqual(data["url"], "http://nohost/plone/test.xml")
        self.assertTrue(data["UID"] is not None)
        # clean it up...
        self.portal.manage_delObjects(["test.xml"])
        transaction.commit()

    def testFileUploadTxt(self):
        view = FileUploadView(self.portal, self.request)
        from plone.namedfile.file import FileChunk

        chunk = FileChunk(b"foobar")
        chunk.filename = "test.txt"
        self.request.form["file"] = chunk
        self.request.REQUEST_METHOD = "POST"
        # the next calls plone.app.dexterity.factories and does a
        # transaction.commit. Needs cleanup and FunctionalTesting layer.
        data = json.loads(view())
        self.assertEqual(data["url"], "http://nohost/plone/test.txt")
        self.assertTrue(data["UID"] is not None)
        # clean it up...
        self.portal.manage_delObjects(["test.txt"])
        transaction.commit()



from Acquisition import aq_inner
from plone.base import PloneMessageFactory as _
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from zope.component import getMultiAdapter
from zope.component import getUtility


class DefaultViewSelectionView(BrowserView):
    def isValidTemplate(self, templateId):
        return templateId in [a[0] for a in self.vocab]

    def canSelectDefaultPage(self):
        return self.context.isPrincipiaFolderish and self.context.canSetDefaultPage()

    @property
    def vocab(self):
        return self.context.getAvailableLayouts()

    @property
    def selectedLayout(self):
        if not self.context_state.is_default_page():
            return self.context.getLayout()
        return ""

    def selectViewTemplate(self):
        templateId = self.request.get("templateId")

        if self.isValidTemplate(templateId):
            self.context.setLayout(templateId)

        self.request.response.redirect(self.context.absolute_url() + "/view")

    @property
    def action_url(self):
        return f"{self.context_state.object_url():s}/select_default_view"

    def __call__(self):
        self.context_state = getMultiAdapter(
            (self.context, self.request), name="plone_context_state"
        )

        template_id = self.request.form.get("templateId")
        context_view_url = self.context_state.object_url() + "/view"

        if self.request.form.get("form.button.Cancel"):
            self.request.response.redirect(context_view_url)

        if self.request.form.get("form.button.Save") and not template_id:
            IStatusMessage(self.request).add(
                "Please select a template.", type="warning"
            )

        if self.request.form.get("form.button.Save") and template_id:
            # Make sure this is a valid template
            if not self.isValidTemplate(template_id):
                IStatusMessage(self.request).add("Invalid view.", type="error")
                return self.index()
            # Update the template
            self.context.setLayout(template_id)
            IStatusMessage(self.request).add("View changed.")
            # Redirect to context view
            self.request.response.redirect(context_view_url)

        return self.index()


class DefaultPageSelectionView(BrowserView):
    def __call__(self):
        if "form.buttons.Save" in self.request.form:
            if "objectId" not in self.request.form:
                message = _("Please select an item to use.")
                msgtype = "error"
            else:
                objectId = self.request.form["objectId"]

                if objectId not in self.context.objectIds():
                    message = _(
                        "There is no object with short name ${name} " "in this folder.",
                        mapping={"name": objectId},
                    )
                    msgtype = "error"
                else:
                    self.context.setDefaultPage(objectId)
                    message = _("View changed.")
                    msgtype = "info"
                    self.request.response.redirect(self.context.absolute_url())
            IStatusMessage(self.request).add(message, msgtype)
        elif "form.buttons.Cancel" in self.request.form:
            self.request.response.redirect(self.context.absolute_url())

        return self.index()

    def get_selectable_items(self):
        """Return brains in this container that can be used as default_pages"""
        context = aq_inner(self.context)
        registry = getUtility(IRegistry)
        view_types = registry.get("plone.types_use_view_action_in_listings", [])
        default_page_types = registry.get("plone.default_page_types", [])
        portal_types = getToolByName(self.context, "portal_types")

        results = []
        for brain in context.getFolderContents():
            portal_type = brain.portal_type
            if portal_type in view_types:
                # Skip files and images
                continue

            if portal_type in default_page_types:
                # Allow types that are explicitly in default_page_types
                results.append(brain)
                continue

            if brain.is_folderish:
                fti = portal_types.get(portal_type)
                if not fti:
                    continue
                if (
                    fti.filter_content_types
                    and fti.allowed_content_types
                    or not fti.filter_content_types
                ):
                    # Disallow folderish types if you can't add any content.
                    # To override you have to add type to default_page_types
                    continue

                results.append(brain)
        return results



from AccessControl import Unauthorized
from plone.app.content.browser.contents import ContentsBaseAction
from plone.app.content.interfaces import IStructureAction
from plone.base import PloneMessageFactory as _
from ZODB.POSException import ConflictError
from zope.i18n import translate
from zope.interface import implementer


@implementer(IStructureAction)
class PasteAction:
    order = 3

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_options(self):
        return {
            "tooltip": translate(_("Paste"), context=self.request),
            "id": "paste",
            "icon": "plone-paste",
            "url": self.context.absolute_url() + "/@@fc-paste",
        }


class PasteActionView(ContentsBaseAction):
    required_obj_permission = "Copy or Move"
    success_msg = _("Successfully pasted items")
    failure_msg = _("Failed to paste items")

    def __call__(self):
        self.protect()
        self.errors = []

        parts = str(self.request.form["folder"].lstrip("/")).split("/")
        parent = self.site.unrestrictedTraverse("/".join(parts[:-1]))
        self.dest = parent.restrictedTraverse(parts[-1])

        try:
            self.dest.manage_pasteObjects(self.request["__cp"])
        except ConflictError:
            raise
        except Unauthorized:
            # avoid this unfriendly exception text:
            # "You are not allowed to access 'manage_pasteObjects' in this
            # context"
            self.errors.append(
                _(
                    "You are not authorized to paste ${title} here.",
                    mapping={"title": self.objectTitle(self.dest)},
                )
            )
        except ValueError as e:
            if "Disallowed subobject type: " in e.args[0]:
                msg_parts = e.args[0].split(":")
                self.errors.append(
                    _(
                        'Disallowed subobject type "${type}"',
                        mapping={"type": msg_parts[1].strip()},
                    )
                )
            else:
                raise e

        return self.message()



from plone.base import PloneMessageFactory as _
from plone.base.utils import transaction_note
from plone.protect import CheckAuthenticator
from plone.protect import PostOnly
from Products.CMFCore.utils import getToolByName
from Products.statusmessages.interfaces import IStatusMessage
from ZODB.POSException import ConflictError
from zope.component import getMultiAdapter
from zope.publisher.browser import BrowserView

import transaction


class FolderPublishView(BrowserView):
    """Publish objects from a folder.

    Originally: Products/CMFPlone/skins/plone_scripts/folder_publish.cpy
    Called by content_status_history, in plone.app.content.
    """

    def __call__(
        self,
        workflow_action=None,
        paths=None,
        comment="No comment",
        expiration_date=None,
        effective_date=None,
        include_children=False,
    ):
        # Use plone.protect.
        PostOnly(self.request)
        CheckAuthenticator(self.request)

        if workflow_action is None:
            IStatusMessage(self.request).add(
                _("You must select a publishing action."), type="error"
            )
            return self.redirect()
        if not paths:
            IStatusMessage(self.request).add(
                _("You must select content to change."), type="error"
            )
            return self.redirect()

        self.transition_objects_by_paths(
            workflow_action,
            paths,
            comment,
            expiration_date,
            effective_date,
            include_children,
        )

        transaction_note(str(paths) + " transitioned " + workflow_action)

        IStatusMessage(self.request).add(_("Item state changed."))
        return self.redirect()

    def transition_objects_by_paths(
        self,
        workflow_action,
        paths,
        comment="",
        expiration_date=None,
        effective_date=None,
        include_children=False,
        handle_errors=True,
    ):
        """Originally this was in plone_utils.transitionObjectsByPaths.

        This was deprecated since 2015, so we copied it here.
        """
        failure = {}
        # use the portal for traversal in case we have relative paths
        portal = getToolByName(self, "portal_url").getPortalObject()
        traverse = portal.restrictedTraverse
        for path in paths:
            sp = transaction.savepoint(optimistic=True)
            try:
                obj = traverse(path, None)
                if obj is not None:
                    view = getMultiAdapter(
                        (obj, self.request), name="content_status_modify"
                    )
                    view(
                        workflow_action,
                        comment,
                        effective_date=effective_date,
                        expiration_date=expiration_date,
                    )
            except (ConflictError, KeyboardInterrupt):
                raise
            except Exception as e:
                # skip this object but continue with sub-objects.
                sp.rollback()
                failure[path] = e
            if getattr(obj, "isPrincipiaFolderish", None) and include_children:
                subobject_paths = [f"{path}/{id}" for id in obj]
                self.transition_objects_by_paths(
                    workflow_action,
                    subobject_paths,
                    comment,
                    expiration_date,
                    effective_date,
                    include_children,
                )
        return failure

    def redirect(self):
        target = self.request.get("orig_template", "")
        if target and not getToolByName(self.context, "portal_url").isURLInPortal(
            target
        ):
            target = ""
        if not target:
            target = self.context.absolute_url()
        self.request.response.redirect(target)



from OFS.interfaces import IOrderedContainer
from plone.app.content.browser.contents import ContentsBaseAction
from plone.app.content.utils import json_loads
from plone.base import PloneMessageFactory as _
from plone.base.interfaces import IPloneSiteRoot
from plone.folder.interfaces import IExplicitOrdering
from Products.CMFCore.utils import getToolByName


class OrderContentsBaseAction(ContentsBaseAction):
    def getOrdering(self):
        if IPloneSiteRoot.providedBy(self.context):
            return self.context
        try:
            if self.context.aq_base.getOrdering():
                ordering = self.context.getOrdering()
            else:
                return None
        except AttributeError:
            if IOrderedContainer.providedBy(self.context):
                # Archetype
                return IOrderedContainer(self.context)
            return None
        if not IExplicitOrdering.providedBy(ordering):
            return None
        return ordering


class ItemOrderActionView(OrderContentsBaseAction):
    success_msg = _("Successfully moved item")
    failure_msg = _("Error moving item")

    def __call__(self):
        self.errors = []
        self.protect()
        id = self.request.form.get("id")
        ordering = self.getOrdering()

        if ordering is None:
            self.errors.append(_("This folder does not support ordering"))
            return self.message()

        delta = self.request.form["delta"]

        if delta == "top":
            ordering.moveObjectsToTop([id])
            return self.message()

        if delta == "bottom":
            ordering.moveObjectsToBottom([id])
            return self.message()

        delta = int(delta)
        subset_ids = json_loads(self.request.form.get("subsetIds", "null"))
        if subset_ids:
            position_id = [(ordering.getObjectPosition(i), i) for i in subset_ids]
            position_id.sort()
            if subset_ids != [i for position, i in position_id]:
                self.errors.append(_("Client/server ordering mismatch"))
                return self.message()

        ordering.moveObjectsByDelta([id], delta, subset_ids)
        return self.message()


class RearrangeActionView(OrderContentsBaseAction):
    success_msg = _("Successfully rearranged folder")
    failure_msg = _("Can not rearrange folder")

    def __call__(self):
        self.protect()
        self.errors = []
        ordering = self.getOrdering()
        if ordering:
            catalog = getToolByName(self.context, "portal_catalog")
            query = {
                "path": {"query": "/".join(self.context.getPhysicalPath()), "depth": 1},
                "sort_on": self.request.form.get("rearrange_on"),
                "show_inactive": True,
            }
            brains = catalog(**query)
            if self.request.form.get("reversed") == "true":
                brains = [b for b in reversed(brains)]
            for idx, brain in enumerate(brains):
                ordering.moveObjectToPosition(brain.id, idx)
        else:
            self.errors.append(_("Not explicit orderable"))
        return self.message()



from AccessControl import Unauthorized
from AccessControl.Permissions import delete_objects
from plone.app.content.browser.contents import ContentsBaseAction
from plone.app.content.interfaces import IStructureAction
from plone.base import PloneMessageFactory as _
from plone.locking.interfaces import ILockable
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import getMultiAdapter
from zope.component.hooks import getSite
from zope.i18n import translate
from zope.interface import implementer

import json


@implementer(IStructureAction)
class DeleteAction:
    template = ViewPageTemplateFile("templates/delete.pt")
    order = 4

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_options(self):
        return {
            "tooltip": translate(_("Delete"), context=self.request),
            "id": "delete",
            "icon": "plone-delete",
            "context": "danger",
            "url": self.context.absolute_url() + "/@@fc-delete",
            "form": {
                "title": translate(_("Delete selected items"), context=self.request),
                "submitText": translate(_("Yes"), context=self.request),
                "submitContext": "danger",
                "template": self.template(),
                "closeText": translate(_("No"), context=self.request),
                "dataUrl": self.context.absolute_url() + "/@@fc-delete",
            },
        }


class DeleteActionView(ContentsBaseAction):
    required_obj_permission = delete_objects
    success_msg = _("Successfully delete items")
    failure_msg = _("Failed to delete items")

    def __call__(self):
        if self.request.form.get("render") == "yes":
            confirm_view = getMultiAdapter(
                (getSite(), self.request), name="delete_confirmation_info"
            )
            selection = self.get_selection()
            catalog = getToolByName(self.context, "portal_catalog")
            brains = catalog(UID=selection, show_inactive=True)
            items = [i.getObject() for i in brains]
            self.request.response.setHeader(
                "Content-Type", "application/json; charset=utf-8"
            )
            return json.dumps({"html": confirm_view(items)})
        else:
            return super().__call__()

    def action(self, obj):
        parent = obj.aq_inner.aq_parent
        title = self.objectTitle(obj)

        try:
            lock_info = obj.restrictedTraverse("@@plone_lock_info")
        except AttributeError:
            lock_info = None
        if lock_info is not None:
            if lock_info.is_locked_for_current_user():
                self.errors.append(
                    _(
                        "${title} is locked and cannot be deleted.",
                        mapping={"title": title},
                    )
                )
                return
            elif lock_info.is_locked():
                # unlock object as it is locked by current user
                ILockable(obj).unlock()

        try:
            parent.manage_delObjects(obj.getId())
        except Unauthorized:
            self.errors.append(
                _(
                    "You are not authorized to delete ${title}.",
                    mapping={"title": self.objectTitle(self.dest)},
                )
            )



from plone.app.content.testing import PLONE_APP_CONTENT_DX_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_PASSWORD
from plone.testing.zope import Browser
from Products.CMFCore.utils import getToolByName

import transaction
import unittest


class ReviewListTestCase(unittest.TestCase):
    """dsfsdaf"""

    layer = PLONE_APP_CONTENT_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.uf = self.portal.acl_users
        self.uf.userFolderAddUser("reviewer", TEST_USER_PASSWORD, ["Reviewer"], [])
        transaction.commit()
        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = True
        self.wftool = getToolByName(self.portal, "portal_workflow")

    def createDocument(self, id, title, description):
        setRoles(
            self.portal,
            TEST_USER_ID,
            [
                "Manager",
            ],
        )
        self.portal.invokeFactory(id=id, type_name="Document")
        doc = getattr(self.portal, id)
        doc.title = title
        doc.description = description
        # we don't want it in the navigation
        doc.exclude_from_nav = True
        doc.reindexObject()
        transaction.commit()
        return doc

    def submitToReview(self, obj):
        """call the workflow action 'submit' for an object"""
        self.wftool.doActionFor(obj, "submit")

    def test_unauthenticated(self):
        """
        unauthenticated users do not have the necessary permissions to view
        the review list
        """
        self.browser.open("http://nohost/plone/full_review_list")
        self.assertTrue("Login Name" in self.browser.contents)

    def test_authenticated(self):
        """
        authenticated users do have the necessary permissions to view
        the review list
        """
        self.browser.addHeader(
            "Authorization", "Basic {}:{}".format("reviewer", TEST_USER_PASSWORD)
        )
        self.browser.open("http://nohost/plone/full_review_list")
        self.assertTrue("Full review list:" in self.browser.contents)

    def test_with_content(self):
        """
        authenticated users do have the necessary permissions to view
        the review list
        """
        doc = self.createDocument("testdoc", "Test Document", "Test Description")
        self.wftool.doActionFor(doc, "submit")
        transaction.commit()

        self.browser.addHeader(
            "Authorization", "Basic {}:{}".format("reviewer", TEST_USER_PASSWORD)
        )
        self.browser.open("http://nohost/plone/full_review_list")
        self.assertTrue("Full review list:" in self.browser.contents)
        # test if the table with review items contains an entry for testdoc
        self.assertTrue('value="/plone/testdoc"' in self.browser.contents)



from setuptools import find_packages
from setuptools import setup


version = "4.1.2"

setup(
    name="plone.app.content",
    version=version,
    description="Content Views for Plone",
    long_description="\n\n".join(
        [
            open("README.rst").read(),
            open("CHANGES.rst").read(),
        ]
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="plone content views viewlet",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.app.content",
    license="GPL version 2",
    packages=find_packages(),
    namespace_packages=["plone", "plone.app"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    extras_require=dict(
        test=[
            "plone.app.contenttypes[test]",
            "plone.app.testing",
            "plone.dexterity",
            "plone.namedfile",
            "plone.testing",
            "zope.annotation",
            "Products.GenericSetup",
        ]
    ),
    install_requires=[
        "plone.app.dexterity",
        "plone.app.querystring",
        "plone.app.uuid",
        "plone.app.vocabularies>4.1.2",
        "plone.app.z3cform",
        "plone.autoform",
        "plone.base",
        "plone.i18n",
        "plone.folder",
        "plone.locking",
        "plone.memoize",
        "plone.protect",
        "plone.supermodel",
        "plone.uuid",
        "Missing",
        "Products.MimetypesRegistry",
        "Products.PortalTransforms",
        "Products.statusmessages",
        "simplejson",
        "z3c.relationfield",
        "z3c.form",
        "zope.browsermenu",
    ],
)



from plone.app.content.browser.constraintypes import IConstrainForm
from zope.interface.exceptions import Invalid

import unittest


class DocumentIntegrationTest(unittest.TestCase):
    def test_formschemainvariants(self):
        class Data:
            allowed_types = []
            secondary_types = []

        bad = Data()
        bad.allowed_types = []
        bad.secondary_types = ["1"]
        good = Data()
        good.allowed_types = ["1"]
        good.secondary_types = []
        self.assertTrue(IConstrainForm.validateInvariants(good) is None)
        self.assertRaises(Invalid, IConstrainForm.validateInvariants, bad)



from OFS.CopySupport import _cb_encode
from OFS.CopySupport import cookie_path
from OFS.Moniker import Moniker
from plone.app.content.browser.contents import ContentsBaseAction
from plone.app.content.interfaces import IStructureAction
from plone.base import PloneMessageFactory as _
from plone.locking.interfaces import ILockable
from zope.i18n import translate
from zope.interface import implementer


@implementer(IStructureAction)
class CutAction:
    order = 1

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_options(self):
        return {
            "tooltip": translate(_("Cut"), context=self.request),
            "id": "cut",
            "icon": "plone-cut",
            "url": self.context.absolute_url() + "/@@fc-cut",
        }


class CutActionView(ContentsBaseAction):
    success_msg = _("Successfully cut items")
    failure_msg = _("Failed to cut items")

    def action(self, obj):
        self.oblist.append(obj)

    def finish(self):
        oblist = []
        for ob in self.oblist:
            try:
                lock_info = ob.restrictedTraverse("@@plone_lock_info")
            except AttributeError:
                lock_info = None
            if lock_info is not None:
                if lock_info.is_locked_for_current_user():
                    self.errors.append(
                        _(
                            "${title} is being edited and cannot be cut.",
                            mapping={"title": self.objectTitle(ob)},
                        )
                    )
                    continue
                elif lock_info.is_locked():
                    # unlock object as it is locked by current user
                    ILockable(ob).unlock()

            if not ob.cb_isMoveable():
                self.errors.append(
                    _(
                        "${title} is being edited and cannot be cut.",
                        mapping={"title": self.objectTitle(ob)},
                    )
                )
                continue
            m = Moniker(ob)
            oblist.append(m.dump())
        cp = (1, oblist)
        cp = _cb_encode(cp)
        resp = self.request.response
        resp.setCookie("__cp", cp, path="%s" % cookie_path(self.request))
        self.request["__cp"] = cp

    def __call__(self):
        self.oblist = []
        return super().__call__()
